from syndot.utils.colors import Color


def print_action(action_type: str, system_target_path: str, settings_target_path: str) -> None:
    if action_type == 'link':
        action = 'Linking'
        preposition = 'to'
        print(f"{action} {Color.link(system_target_path)} {preposition} {Color.settings(settings_target_path)}")
    elif action_type == 'unlink':
        action = 'Unlinking'
        preposition = 'from'
        print(f"{action} {Color.link(system_target_path)} {preposition} {Color.settings(settings_target_path)}")
    elif action_type == 'diffuse':
        action = 'Diffusing'
        preposition = 'to'
        print(f"{action} {Color.settings(settings_target_path)} {preposition} {Color.link(system_target_path)}")


def print_relationship(system_target_path: str, settings_target_path: str, symbol: str) -> None:
    print(f"{Color.link(system_target_path)} {Color.symbol(symbol)} {Color.settings(settings_target_path)}")


def print_highlight(sentence: str, end: str = '\n') -> None:
    print(f"{Color.highlight(sentence)}", end = end)


def print_error(error: str, end: str = '\n') -> None:
    print(f"{Color.error(error)}", end = end)



from argparse import SUPPRESS
from syndot.parser.parser import command_parser, CommandFormatter


init_parser = command_parser.add_parser('init',
                                        prog = 'syndot init',
                                        usage = '%(prog)s [OPTION]',
                                        description = 'Initialize <SETTINGS> directory where store dotfiles',
                                        help = 'Initialize <SETTINGS> directory where store dotfiles',
                                        add_help = False,
                                        formatter_class = CommandFormatter)
init_parser._positionals.title = 'ARGUMENTS'
init_parser._optionals.title = 'OPTIONS'
init_parser.add_argument('-h', '--help',
                         action = 'help',
                         default = SUPPRESS,
                         help = 'Show this help message and exit')
init_parser.add_argument('-p', '--path',
                         required = False,
                         help = 'Path to the <SETTINGS> directory. If not provided, create the <SETTINGS> directory at '
                                'the default path: ~/Settings')



from argparse import Namespace
from configparser import ConfigParser
import os
from syndot.utils.path import expand_home_path


def read_map_file(map_file_path: str | None) -> ConfigParser:
    map_file_path = expand_home_path(map_file_path if map_file_path is not None else 'map.ini')
    if not os.path.exists(map_file_path):
        if map_file_path == 'map.ini':
            raise FileNotFoundError("Missing map.ini file in current directory.")
        else:
            raise FileNotFoundError("Missing map.ini file at the specified path.")
    config = ConfigParser()
    config.read(map_file_path)

    return config


def get_map_info(config: ConfigParser, args: Namespace) -> tuple[str, list[str]]:
    settings_dir = config['Path']['settings_dir']
    target_directories = config['Targets']['directories'].split()
    target_files = config['Targets']['files'].split()
    if args.TARGET_PATH_START is None and args.all:
        targets = [*target_files, *target_directories]
    else:
        targets = []
        for file in target_files:
            if args.exact:
                if file == args.TARGET_PATH_START:
                    targets.append(file)
            else:
                if file.startswith(args.TARGET_PATH_START):
                    targets.append(file)
        for directory in target_directories:
            if args.exact:
                if directory == args.TARGET_PATH_START:
                    targets.append(directory)
            else:
                if directory.startswith(args.TARGET_PATH_START):
                    targets.append(directory)

    return settings_dir, targets


def write_map_file(map_file_path: str | None, config: ConfigParser) -> None:
    with open(map_file_path, 'w') as map_file:
        config.write(map_file)



from syndot import commands
from syndot.parser.parser import parser
from syndot import init_config


init_config.init_config()
args = parser.parse_args()

command_map = {'init': commands.init,
               'link': commands.link,
               'unlink': commands.unlink,
               'diffuse': commands.diffuse,
               'add': commands.add,
               'remove': commands.remove,
               'list': commands.list_}

command_map[args.command](args = args)



from argparse import SUPPRESS
from syndot.parser.parser import command_parser, CommandFormatter


add_parser = command_parser.add_parser('add',
                                       prog = 'syndot add',
                                       usage = '%(prog)s [OPTION] ARGUMENT',
                                       description = 'Add dotfiles to the map file',
                                       help = 'Add dotfiles to the map file',
                                       add_help = False,
                                       formatter_class = CommandFormatter)
add_parser._positionals.title = 'ARGUMENTS'
add_parser._optionals.title = 'OPTIONS'
add_parser.add_argument('TARGET_PATH',
                        help = 'Dotfile path')
add_parser.add_argument('-h', '--help',
                        action = 'help',
                        default = SUPPRESS,
                        help = 'Show this help message and exit')
add_parser.add_argument('-m', '--mapfile',
                        required = False,
                        metavar = '<MAP_FILE>',
                        help = 'Path to the %(metavar)s. If not provided search for a \'map.ini\' file in the '
                               'current directory, so not required if current directory is the <SETTINGS> directory')



from argparse import Namespace
import os
import shutil
from syndot.init_config import LOG_FILE_PATH
from syndot.utils.commands import skip_dotfiles
from syndot.utils.file_actions import remove
from syndot.utils.logger import log_error
from syndot.utils.map_file import get_map_info, read_map_file
from syndot.utils.path import compose_target_paths, expand_home_path, generate_backup_path
from syndot.utils.print_ import print_action, print_error, print_highlight, print_relationship
from syndot.utils.prompt import ask_to_proceed


def unlink(args: Namespace) -> None:
    settings_dir, targets = get_map_info(config = read_map_file(map_file_path = args.mapfile), args = args)

    targets_to_be_unlinked = {}
    wrong_existing_links = {}
    already_existing_system = {}
    already_unlinked_targets = []
    missing_settings_targets = []
    settings_are_links = []

    print_highlight('Looking for files and directories to unlink...')

    for target in targets:
        system_target_path, settings_target_path = compose_target_paths(settings_dir = settings_dir, target = target)

        if not os.path.islink(settings_target_path):
            if os.path.exists(settings_target_path):
                if os.path.islink(system_target_path):
                    if os.readlink(system_target_path) == settings_target_path:
                        targets_to_be_unlinked[system_target_path] = settings_target_path
                    else:
                        wrong_existing_links[system_target_path] = settings_target_path
                else:
                    if not os.path.exists(system_target_path):
                        targets_to_be_unlinked[system_target_path] = settings_target_path
                    else:
                        already_existing_system[system_target_path] = settings_target_path
            else:
                if os.path.exists(system_target_path):
                    already_unlinked_targets.append(settings_target_path)
                else:
                    missing_settings_targets.append(settings_target_path)
        else:
            settings_are_links.append(settings_target_path)

    if not any([targets_to_be_unlinked, wrong_existing_links, already_existing_system, already_unlinked_targets,
                missing_settings_targets, settings_are_links]):
        print_highlight('No files or directories found to unlink.')

    unlink_dotfiles(targets_list = targets_to_be_unlinked,
                    settings_dir = settings_dir,
                    many_targets_sentence = f"Found {len(targets_to_be_unlinked.keys())} files or directories to "
                                            f"unlink.\nThey will be moved to the respective system directories.\nThe "
                                            f"original symbolic links and eventually the backups will be removed.",
                    single_file_sentence = f"Found {len(targets_to_be_unlinked.keys())} file to unlink.\nIt will be "
                                           f"moved to the respective system directory.\nThe original symbolic link and "
                                           f"eventually the backup file will be removed.",
                    single_directory_sentence = f"Found {len(targets_to_be_unlinked.keys())} directory to unlink.\n"
                                                f"It will be moved to the respective system directory.\nThe original "
                                                f"symbolic link and eventually the backup directory will be removed.")

    unlink_dotfiles(targets_list = wrong_existing_links,
                    settings_dir = settings_dir,
                    many_targets_sentence = f"Found {len(wrong_existing_links.keys())} files or directories that are "
                                            f"links to wrong files or directories.\nThese wrong links will be removed "
                                            f"and eventually also the backups.\nThen the settings files and "
                                            f"directories will be moved to the respective system directories.",
                    single_file_sentence = f"Found {len(wrong_existing_links.keys())} file that is a link to a wrong "
                                           f"file.\nThis wrong link will be removed and eventually also the backup "
                                           f"file.\nThen the settings file will be moved to the respective system "
                                           f"directory.",
                    single_directory_sentence = f"Found {len(wrong_existing_links.keys())} directory that is a link to "
                                                f"a wrong directory.\nThis wrong link will be removed and eventually "
                                                f"also the backup directory.\nThen the settings directory will be "
                                                f"moved to the respective system directory.")

    unlink_dotfiles(targets_list = already_existing_system,
                    settings_dir = settings_dir,
                    many_targets_sentence = f"Found {len(already_existing_system.keys())} already existing system "
                                            f"files or directories.\nThey will be removed and eventually also the "
                                            f"backups.\nThen the settings files and directories will be moved to the "
                                            f"respective system directories.",
                    single_file_sentence = f"Found {len(already_existing_system.keys())} already existing system "
                                           f"file.\nIt will be removed and eventually also the backup file.\nThen the "
                                           f"settings file will be moved to the respective system directory.",
                    single_directory_sentence = f"Found {len(already_existing_system.keys())} already existing system "
                                                f"directory.\nIt will be removed and eventually also the backup "
                                                f"directory.\nThen the settings directory will be moved to the "
                                                f"respective system directory.")

    skip_dotfiles(targets_list = already_unlinked_targets,
                  many_targets_sentence = f"Skipping {len(already_unlinked_targets)} already unlinked files or "
                                          f"directories.",
                  single_file_sentence = f"Skipping {len(already_unlinked_targets)} already unlinked file.",
                  single_directory_sentence = f"Skipping {len(already_unlinked_targets)} already unlinked directory.")

    skip_dotfiles(targets_list = missing_settings_targets,
                  many_targets_sentence = f"Skipping {len(missing_settings_targets)} missing settings files or "
                                          f"directories.",
                  single_file_sentence = f"Skipping {len(missing_settings_targets)} missing settings file.",
                  single_directory_sentence = f"Skipping {len(missing_settings_targets)} missing settings directory.")

    skip_dotfiles(targets_list = settings_are_links,
                  many_targets_sentence = f"Skipping {len(settings_are_links)} settings files or directories because "
                                          f"are links.",
                  single_file_sentence = f"Skipping {len(settings_are_links)} settings file because is a link.",
                  single_directory_sentence = f"Skipping {len(settings_are_links)} settings directory because is a "
                                              f"link.")


def unlink_dotfiles(targets_list: dict[str, str],
                    settings_dir: str,
                    many_targets_sentence: str,
                    single_file_sentence: str,
                    single_directory_sentence: str) -> None:
    if targets_list:
        n_targets = len(targets_list.keys())
        if n_targets > 1:
            for system_target_path, settings_target_path in targets_list.items():
                print_relationship(system_target_path = system_target_path,
                                   settings_target_path = settings_target_path,
                                   symbol = '-x->')
            print_highlight(many_targets_sentence)
            proceed = ask_to_proceed()
        else:
            if os.path.isfile(list(targets_list.keys())[0]):
                print_relationship(system_target_path = list(targets_list.keys())[0],
                                   settings_target_path = list(targets_list.values())[0],
                                   symbol = '-x->')
                print_highlight(single_file_sentence)
                proceed = ask_to_proceed()
            else:
                print_relationship(system_target_path = list(targets_list.keys())[0],
                                   settings_target_path = list(targets_list.values())[0],
                                   symbol = '-x->')
                print_highlight(single_directory_sentence)
                proceed = ask_to_proceed()

        if proceed:
            for i, (system_target_path, settings_target_path) in enumerate(targets_list.items(), 1):
                try:
                    print_action(action_type = 'unlink',
                                 system_target_path = system_target_path,
                                 settings_target_path = settings_target_path)
                    print(f"Total ({i}/{n_targets})", end = '\r')

                    remove(path = system_target_path)
                    system_target_parent = os.path.dirname(system_target_path)
                    if not os.path.exists(system_target_parent):
                        os.makedirs(system_target_parent)
                    shutil.move(settings_target_path, system_target_path)
                    backup_path = generate_backup_path(path = system_target_path)
                    remove(path = backup_path)

                    parent_directory = os.path.dirname(settings_target_path)
                    protected_directories = (settings_dir, expand_home_path('~'), '/')
                    while parent_directory not in protected_directories:
                        parent_directory_content = os.listdir(parent_directory)
                        if not parent_directory_content or ((len(parent_directory_content) == 1) and
                                                            parent_directory_content[0] == '.directory'):
                            shutil.rmtree(parent_directory)
                            parent_directory = os.path.dirname(parent_directory)
                        else:
                            break
                except Exception as e:
                    print_error(f"Error in unlinking {system_target_path}. Check {LOG_FILE_PATH} for more details.")
                    log_error(f'{e}')
        print('\n')



from argparse import Namespace
import os
from syndot.utils.map_file import read_map_file, write_map_file
from syndot.utils.path import expand_home_path


def remove(args: Namespace) -> None:
    map_file_path = expand_home_path(args.mapfile if args.mapfile is not None else 'map.ini')

    config = read_map_file(map_file_path = map_file_path)
    current_files = config['Targets']['files'].split()
    current_directories = config['Targets']['directories'].split()
    target = args.TARGET_PATH
    if target.endswith(os.sep):
        target = target[:-1]
    if target in current_files:
        current_files.remove(target)
        config['Targets']['files'] = '\n' + '\n'.join(current_files)
    elif target in current_directories:
        current_directories.remove(target)
        config['Targets']['directories'] = '\n' + '\n'.join(current_directories)
    else:
        raise NameError(f"Target {target} not found in map file")

    write_map_file(map_file_path = map_file_path, config = config)



VALID_PROMPT_CHOICES = {'y': True, 'ye': True, 'yes': True, 'n': False, 'no': False}


def ask_to_proceed() -> bool:
    choice = ''
    while choice not in VALID_PROMPT_CHOICES:
        choice = input("Do you want to proceed (y/N)? ").lower()
        if choice == '':
            choice = 'n'

    return VALID_PROMPT_CHOICES[choice]



from argparse import SUPPRESS
from syndot.parser.parser import command_parser, CommandFormatter


list_parser = command_parser.add_parser('list',
                                        prog = 'syndot list',
                                        usage = '%(prog)s [OPTION]',
                                        description = 'List dotfiles in the map file',
                                        help = 'List dotfiles in the map file',
                                        add_help = False,
                                        formatter_class = CommandFormatter)
list_parser._positionals.title = 'ARGUMENTS'
list_parser._optionals.title = 'OPTIONS'
list_parser.add_argument('-h', '--help',
                         action = 'help',
                         default = SUPPRESS,
                         help = 'Show this help message and exit')
list_parser.add_argument('-m', '--mapfile',
                         required = False,
                         metavar = '<MAP_FILE>',
                         help = 'Path to the %(metavar)s. If not provided search for a \'map.ini\' file in the '
                                'current directory, so not required if current directory is the <SETTINGS> directory')



from argparse import SUPPRESS
from syndot.parser.parser import command_parser, CommandFormatter


remove_parser = command_parser.add_parser('remove',
                                          prog = 'syndot remove',
                                          usage = '%(prog)s [OPTION] ARGUMENT',
                                          description = 'Remove dotfiles from the map file',
                                          help = 'Remove dotfiles from the map file',
                                          add_help = False,
                                          formatter_class = CommandFormatter)
remove_parser._positionals.title = 'ARGUMENTS'
remove_parser._optionals.title = 'OPTIONS'
remove_parser.add_argument('TARGET_PATH',
                           help = 'Dotfile path')
remove_parser.add_argument('-h', '--help',
                           action = 'help',
                           default = SUPPRESS,
                           help = 'Show this help message and exit')
remove_parser.add_argument('-m', '--mapfile',
                           required = False,
                           metavar = '<MAP_FILE>',
                           help = 'Path to the %(metavar)s. If not provided search for a \'map.ini\' file in the '
                                  'current directory, so not required if current directory is the <SETTINGS> directory')



import os
from syndot.utils.print_ import print_highlight


def skip_dotfiles(targets_list: list[str],
                  many_targets_sentence: str,
                  single_file_sentence: str,
                  single_directory_sentence: str) -> None:
    if targets_list:
        for target in targets_list:
            print(f"Skip {target}")
        if len(targets_list) > 1:
            print_highlight(many_targets_sentence)
        else:
            if os.path.isfile(targets_list[0]):
                print_highlight(single_file_sentence)
            else:
                print_highlight(single_directory_sentence)
        print()



from argparse import SUPPRESS
from syndot.parser.parser import command_parser, CommandFormatter


unlink_parser = command_parser.add_parser('unlink',
                                          prog = 'syndot unlink',
                                          usage = '%(prog)s [OPTION] [ARGUMENT]',
                                          description = 'Remove symlinks and move dotfiles from <SETTINGS> directory '
                                                        'to their original directories',
                                          help = 'Remove symlinks and move dotfiles from <SETTINGS> directory to their '
                                                 'original directories',
                                          add_help = False,
                                          formatter_class = CommandFormatter)
unlink_parser._positionals.title = 'ARGUMENTS'
unlink_parser._optionals.title = 'OPTIONS'
unlink_targets = unlink_parser.add_mutually_exclusive_group(required = True)
unlink_targets.add_argument('TARGET_PATH_START',
                            nargs = '?',
                            metavar = 'TARGET_PATH',
                            help = 'Dotfile path, mandatory if [-a | -all] option is not provided')
unlink_targets.add_argument('-a', '--all',
                            action = 'store_true',
                            default = False,
                            required = False,
                            dest = 'all',
                            help = 'Select all dotfiles in the <MAP_FILE>, mandatory if <TARGET_PATH> argument is not '
                                   'provided')
unlink_parser.add_argument('-e', '--exact',
                           action = 'store_true',
                           default = False,
                           required = False,
                           dest = 'exact',
                           help = 'Search for an exact match for the <TARGET_PATH> in the map file. If not provided, '
                                  'search for all targets which paths begin with <TARGET_PATH> in the map file')
unlink_parser.add_argument('-h', '--help',
                           action = 'help',
                           default = SUPPRESS,
                           help = 'Show this help message and exit')
unlink_parser.add_argument('-m', '--mapfile',
                           required = False,
                           metavar = '<MAP_FILE>',
                           help = 'Path to the %(metavar)s. If not provided search for a \'map.ini\' file in the '
                                  'current directory, so not required if current directory is the <SETTINGS> directory')



from argparse import Namespace
from syndot.utils.map_file import read_map_file
from syndot.utils.path import expand_home_path
from syndot.utils.print_ import print_highlight
from syndot.utils.colors import Color


def list_(args: Namespace) -> None:
    config = read_map_file(map_file_path = args.mapfile)

    settings_dir = expand_home_path(config['Path']['settings_dir'])
    target_directories = config['Targets']['directories'].split()
    target_files = config['Targets']['files'].split()
    targets = [*target_directories, *target_files]
    targets.sort()

    for target in targets:
        print(target)
    print_highlight("Settings directory:", end = ' ')
    print(Color.settings(settings_dir))



from argparse import Namespace
import os
from syndot.utils.map_file import read_map_file, write_map_file
from syndot.utils.path import expand_home_path


def add(args: Namespace) -> None:
    map_file_path = expand_home_path(args.mapfile if args.mapfile is not None else 'map.ini')

    target = args.TARGET_PATH
    if target.endswith(os.sep):
        target = target[:-1]
    if not os.path.exists(target):
        raise OSError(f"Target {target} not found")

    config = read_map_file(map_file_path = map_file_path)
    current_targets = []
    if os.path.isfile(target):
        current_targets = config['Targets']['files'].split()
    elif os.path.isdir(target):
        current_targets = config['Targets']['directories'].split()

    target_path = expand_home_path(target)
    if target_path in current_targets:
        print(f"Target {target} already in map file")
        return
    current_targets.append(target_path)
    current_targets = list(set(current_targets))
    current_targets.sort()

    if os.path.isfile(target):
        config['Targets']['files'] = '\n' + '\n'.join(current_targets)
    elif os.path.isdir(target):
        config['Targets']['directories'] = '\n' + '\n'.join(current_targets)

    write_map_file(map_file_path = map_file_path, config = config)



from argparse import SUPPRESS
from syndot.parser.parser import command_parser, CommandFormatter


link_parser = command_parser.add_parser('link',
                                        prog = 'syndot link',
                                        usage = '%(prog)s [OPTION] [ARGUMENT]',
                                        description = 'Move dotfiles to <SETTINGS> directory and create symlinks to '
                                                      'them',
                                        help = 'Move dotfiles to <SETTINGS> directory and create symlinks to them',
                                        add_help = False,
                                        formatter_class = CommandFormatter)
link_parser._positionals.title = 'ARGUMENTS'
link_parser._optionals.title = 'OPTIONS'
link_targets = link_parser.add_mutually_exclusive_group(required = True)
link_targets.add_argument('TARGET_PATH_START',
                          nargs = '?',
                          metavar = 'TARGET_PATH',
                          help = 'Dotfile path, mandatory if [-a | -all] option is not provided')
link_targets.add_argument('-a', '--all',
                          action = 'store_true',
                          default = False,
                          required = False,
                          dest = 'all',
                          help = 'Select all dotfiles in the <MAP_FILE>, mandatory if <TARGET_PATH> argument is not '
                                 'provided')
link_parser.add_argument('-b', '--backup',
                         action = 'store_true',
                         default = False,
                         required = False,
                         dest = 'backup',
                         help = 'Create a backup copy of the original dotfile')
link_parser.add_argument('-e', '--exact',
                         action = 'store_true',
                         default = False,
                         required = False,
                         dest = 'exact',
                         help = 'Search for an exact match for the <TARGET_PATH> in the map file. If not provided, '
                                'search for all targets which paths begin with <TARGET_PATH> in the map file')
link_parser.add_argument('-h', '--help',
                         action = 'help',
                         default = SUPPRESS,
                         help = 'Show this help message and exit')
link_parser.add_argument('-m', '--mapfile',
                         required = False,
                         metavar = '<MAP_FILE>',
                         help = 'Path to the %(metavar)s. If not provided search for a \'map.ini\' file in the current '
                                'directory, so not required if current directory is the <SETTINGS> directory')



from hypothesis import given, settings, HealthCheck
from hypothesis.strategies import one_of, none, booleans, text, characters, sampled_from
from pytest import mark, raises
from syndot.parser.parser import parser
from tests.conftest import paths


@mark.utils
class TestParser:

    @mark.genuine
    @given(path = one_of(paths(), none()))
    @settings(max_examples = 100, deadline = None)
    def test_init(self, path):
        input_arguments = ['init']
        if path:
            input_arguments.extend(['-p', path])

        parsed_arguments = parser.parse_args(input_arguments)

        assert parsed_arguments.command == 'init'
        if path:
            assert parsed_arguments.path == path
        else:
            assert parsed_arguments.path is None

    @mark.genuine
    @given(backup = booleans(),
           exact = booleans(),
           map_file_path = one_of(paths(), none()),
           target_path_start = one_of(text(min_size = 5, max_size = 10, alphabet = characters(min_codepoint = 97, max_codepoint = 122)),
                                      none()))
    @settings(max_examples = 100, deadline = None)
    def test_link(self, backup, exact, map_file_path, target_path_start):
        input_arguments = ['link']
        if backup:
            input_arguments.append('-b')
        if exact:
            input_arguments.append('-e')
        if map_file_path:
            input_arguments.extend(['-m', map_file_path])
        if target_path_start:
            input_arguments.append(target_path_start)
        else:
            input_arguments.append('--all')

        parsed_arguments = parser.parse_args(input_arguments)

        assert parsed_arguments.command == 'link'
        if backup:
            assert parsed_arguments.backup
        else:
            assert not parsed_arguments.backup
        if exact:
            assert parsed_arguments.exact
        else:
            assert not parsed_arguments.exact
        if map_file_path:
            assert parsed_arguments.mapfile == map_file_path
        else:
            assert parsed_arguments.mapfile is None
        if target_path_start:
            assert parsed_arguments.TARGET_PATH_START == target_path_start
            assert not parsed_arguments.all
        else:
            assert parsed_arguments.TARGET_PATH_START is None
            assert parsed_arguments.all

    @mark.genuine
    @given(exact = booleans(),
           map_file_path = one_of(paths(), none()),
           target_path_start = one_of(text(min_size = 5, max_size = 10, alphabet = characters(min_codepoint = 97, max_codepoint = 122)),
                                      none()))
    @settings(max_examples = 100, deadline = None)
    def test_unlink(self, exact, map_file_path, target_path_start):
        input_arguments = ['unlink']
        if exact:
            input_arguments.append('-e')
        if map_file_path:
            input_arguments.extend(['-m', map_file_path])
        if target_path_start:
            input_arguments.append(target_path_start)
        else:
            input_arguments.append('--all')

        parsed_arguments = parser.parse_args(input_arguments)

        assert parsed_arguments.command == 'unlink'
        if exact:
            assert parsed_arguments.exact
        else:
            assert not parsed_arguments.exact
        if map_file_path:
            assert parsed_arguments.mapfile == map_file_path
        else:
            assert parsed_arguments.mapfile is None
        if target_path_start:
            assert parsed_arguments.TARGET_PATH_START == target_path_start
            assert not parsed_arguments.all
        else:
            assert parsed_arguments.TARGET_PATH_START is None
            assert parsed_arguments.all

    @mark.genuine
    @given(exact = booleans(),
           map_file_path = one_of(paths(), none()),
           target_path_start = one_of(text(min_size = 5, max_size = 10, alphabet = characters(min_codepoint = 97, max_codepoint = 122)),
                                      none()))
    @settings(max_examples = 100, deadline = None)
    def test_diffuse(self, exact, map_file_path, target_path_start):
        input_arguments = ['diffuse']
        if exact:
            input_arguments.append('-e')
        if map_file_path:
            input_arguments.extend(['-m', map_file_path])
        if target_path_start:
            input_arguments.append(target_path_start)
        else:
            input_arguments.append('--all')

        parsed_arguments = parser.parse_args(input_arguments)

        assert parsed_arguments.command == 'diffuse'
        if exact:
            assert parsed_arguments.exact
        else:
            assert not parsed_arguments.exact
        if map_file_path:
            assert parsed_arguments.mapfile == map_file_path
        else:
            assert parsed_arguments.mapfile is None
        if target_path_start:
            assert parsed_arguments.TARGET_PATH_START == target_path_start
            assert not parsed_arguments.all
        else:
            assert parsed_arguments.TARGET_PATH_START is None
            assert parsed_arguments.all

    @mark.genuine
    @given(map_file_path = one_of(paths(), none()),
           target_path = text(min_size = 5, max_size = 10, alphabet = characters(min_codepoint = 97, max_codepoint = 122)))
    @settings(max_examples = 100, deadline = None)
    def test_add(self, map_file_path, target_path):
        input_arguments = ['add']
        if map_file_path:
            input_arguments.extend(['-m', map_file_path])
        input_arguments.append(target_path)

        parsed_arguments = parser.parse_args(input_arguments)

        assert parsed_arguments.command == 'add'
        if map_file_path:
            assert parsed_arguments.mapfile == map_file_path
        else:
            assert parsed_arguments.mapfile is None
        assert parsed_arguments.TARGET_PATH == target_path

    @mark.genuine
    @given(map_file_path = one_of(paths(), none()),
           target_path = text(min_size = 5, max_size = 10, alphabet = characters(min_codepoint = 97, max_codepoint = 122)))
    @settings(max_examples = 100, deadline = None)
    def test_remove(self, map_file_path, target_path):
        input_arguments = ['remove']
        if map_file_path:
            input_arguments.extend(['-m', map_file_path])
        input_arguments.append(target_path)

        parsed_arguments = parser.parse_args(input_arguments)

        assert parsed_arguments.command == 'remove'
        if map_file_path:
            assert parsed_arguments.mapfile == map_file_path
        else:
            assert parsed_arguments.mapfile is None
        assert parsed_arguments.TARGET_PATH == target_path

    @mark.genuine
    @given(map_file_path = one_of(paths(), none()))
    @settings(max_examples = 100, deadline = None)
    def test_list(self, map_file_path):
        input_arguments = ['list']
        if map_file_path:
            input_arguments.extend(['-m', map_file_path])

        parsed_arguments = parser.parse_args(input_arguments)

        assert parsed_arguments.command == 'list'
        if map_file_path:
            assert parsed_arguments.mapfile == map_file_path
        else:
            assert parsed_arguments.mapfile is None

    @mark.genuine
    @given(abbreviation = booleans())
    @settings(max_examples = 100, deadline = None, suppress_health_check = [HealthCheck.function_scoped_fixture])
    def test_version(self, abbreviation, capsys):
        args = ['-v'] if abbreviation else ['--version']
        with raises(SystemExit):
            parser.parse_args(args = args)
        printed_output = capsys.readouterr().out

        assert printed_output.startswith(f'syndot')

    @mark.genuine
    @given(abbreviation = booleans())
    @settings(max_examples = 100, deadline = None, suppress_health_check = [HealthCheck.function_scoped_fixture])
    def test_help(self, abbreviation, capsys):
        args = ['-h'] if abbreviation else ['--help']
        with raises(SystemExit):
            parser.parse_args(args = args)
        printed_output = capsys.readouterr().out

        assert printed_output.startswith('usage: syndot {OPTION | COMMAND}')
        assert printed_output.endswith('Config file path: ~/.config/syndot\n')
        assert 'OPTIONS' in printed_output
        assert 'COMMANDS' in printed_output
        assert '-h, --help' in printed_output
        assert '-v, --version' in printed_output
        assert 'add' in printed_output
        assert 'diffuse' in printed_output
        assert 'init' in printed_output
        assert 'link' in printed_output
        assert 'list' in printed_output
        assert 'remove' in printed_output
        assert 'unlink' in printed_output

    @mark.genuine
    @given(command = sampled_from(elements = ['add', 'diffuse', 'init', 'link', 'list', 'remove', 'unlink']),
           abbreviation = booleans())
    @settings(max_examples = 100, deadline = None, suppress_health_check = [HealthCheck.function_scoped_fixture])
    def test_command_help(self, command, abbreviation, capsys):
        args = [command]
        if abbreviation:
            args.append('-h')
        else:
            args.append('--help')
        with raises(SystemExit):
            parser.parse_args(args = args)
        printed_output = capsys.readouterr().out

        assert printed_output.startswith(f'usage: syndot {command}')
        assert 'OPTIONS' in printed_output
        if command in ['add', 'diffuse', 'link', 'remove', 'unlink']:
            assert 'ARGUMENTS' in printed_output
        assert '-h, --help' in printed_output
        if command in ['add', 'diffuse', 'link', 'list', 'remove', 'unlink']:
            assert '-m, --mapfile <MAP_FILE>' in printed_output
        if command in ['diffuse', 'link', 'unlink']:
            assert '-a, --all' in printed_output
            assert '-e, --exact' in printed_output
            assert '-e, --exact' in printed_output
        if command == 'init':
            assert '-p, --path PATH' in printed_output
        if command == 'link':
            assert '-b, --backup' in printed_output



from argparse import SUPPRESS
from syndot.parser.parser import command_parser, CommandFormatter


diffuse_parser = command_parser.add_parser('diffuse',
                                           prog = 'syndot diffuse',
                                           usage = '%(prog)s [OPTION] [ARGUMENT]',
                                           description = 'Create dotfiles symlinks',
                                           help = 'Create dotfiles symlinks',
                                           add_help = False,
                                           formatter_class = CommandFormatter)
diffuse_parser._positionals.title = 'ARGUMENTS'
diffuse_parser._optionals.title = 'OPTIONS'
diffuse_targets = diffuse_parser.add_mutually_exclusive_group(required = True)
diffuse_targets.add_argument('TARGET_PATH_START',
                             nargs = '?',
                             metavar = 'TARGET_PATH',
                             help = 'Dotfile path, mandatory if [-a | -all] option is not provided')
diffuse_targets.add_argument('-a', '--all',
                             action = 'store_true',
                             default = False,
                             required = False,
                             dest = 'all',
                             help = 'Select all dotfiles in the <MAP_FILE>, mandatory if <TARGET_PATH> argument is not '
                                    'provided')
diffuse_parser.add_argument('-e', '--exact',
                            action = 'store_true',
                            default = False,
                            required = False,
                            dest = 'exact',
                            help = 'Search for an exact match for the <TARGET_PATH> in the map file. If not provided, '
                                   'search for all targets which paths begin with <TARGET_PATH> in the map file')
diffuse_parser.add_argument('-h', '--help',
                            action = 'help',
                            default = SUPPRESS,
                            help = 'Show this help message and exit')
diffuse_parser.add_argument('-m', '--mapfile',
                            required = False,
                            metavar = '<MAP_FILE>',
                            help = 'Path to the %(metavar)s. If not provided search for a \'map.ini\' file in the '
                                   'current directory, so not required if current directory is the <SETTINGS> directory')



from argparse import Namespace
import os
from syndot.init_config import LOG_FILE_PATH
from syndot.utils.commands import skip_dotfiles
from syndot.utils.file_actions import change_parent_owner, change_child_owner, copy, remove
from syndot.utils.logger import log_error
from syndot.utils.map_file import get_map_info, read_map_file
from syndot.utils.path import compose_target_paths, generate_backup_path
from syndot.utils.print_ import print_action, print_error, print_highlight, print_relationship
from syndot.utils.prompt import ask_to_proceed


def link(args: Namespace) -> None:
    settings_dir, targets = get_map_info(config = read_map_file(map_file_path = args.mapfile), args = args)

    targets_to_be_linked = {}
    already_existing_settings = {}
    missing_system_targets = []
    already_linked_targets = []
    corrupted_targets = []
    wrong_existing_links = []

    print_highlight('Looking for files and directories to link...')

    for target in targets:
        system_target_path, settings_target_path = compose_target_paths(settings_dir = settings_dir, target = target)

        if not os.path.islink(system_target_path):
            if os.path.exists(system_target_path):
                if not os.path.exists(settings_target_path):
                    targets_to_be_linked[system_target_path] = settings_target_path
                else:
                    already_existing_settings[system_target_path] = settings_target_path
            else:
                missing_system_targets.append(system_target_path)
        else:
            if os.readlink(system_target_path) == settings_target_path:
                if os.path.exists(system_target_path):
                    already_linked_targets.append(system_target_path)
                else:
                    corrupted_targets.append(system_target_path)
            else:
                wrong_existing_links.append(system_target_path)

    if not any([targets_to_be_linked, already_existing_settings, missing_system_targets, already_linked_targets,
                corrupted_targets, wrong_existing_links]):
        print_highlight('No files or directories found to link.')

    link_dotfiles(targets_list = targets_to_be_linked,
                  settings_dir = settings_dir,
                  backup = args.backup,
                  many_targets_sentence = f"Found {len(targets_to_be_linked.keys())} files or directories to link.\n"
                                          f"They will be moved to settings directory and will be replaced by "
                                          f"symbolic links.",
                  single_file_sentence = f"Found {len(targets_to_be_linked.keys())} file to link.\nIt will be moved to "
                                         f"settings directory and will be replaced by a symbolic link.",
                  single_directory_sentence = f"Found {len(targets_to_be_linked.keys())} directory to link.\nIt will "
                                              f"be moved to settings directory and will be replaced by a symbolic "
                                              f"link.",
                  remove_settings = False)

    link_dotfiles(targets_list = already_existing_settings,
                  settings_dir = settings_dir,
                  backup = args.backup,
                  many_targets_sentence = f"Found {len(already_existing_settings.keys())} already existing files or "
                                          f"directories in settings directory.\nThey will be removed, then the "
                                          f"respective files and directories from the system will be moved to settings "
                                          f"directory and will be replaced by symbolic links.",
                  single_file_sentence = f"Found {len(already_existing_settings.keys())} already existing file in "
                                         f"settings directory.\nIt will be removed, then the respective file from the "
                                         f"system will be moved to settings directory and will be replaced by a "
                                         f"symbolic link.",
                  single_directory_sentence = f"Found {len(already_existing_settings.keys())} already existing "
                                              f"directory in settings directory.\nIt will be removed, then the "
                                              f"respective directory from the system will be moved to settings "
                                              f"directory and will be replaced by a symbolic link.",
                  remove_settings = True)

    skip_dotfiles(targets_list = missing_system_targets,
                  many_targets_sentence = f"Skipping {len(missing_system_targets)} missing files or directories.",
                  single_file_sentence = f"Skipping {len(missing_system_targets)} missing file.",
                  single_directory_sentence = f"Skipping {len(missing_system_targets)} missing directory.")

    skip_dotfiles(targets_list = already_linked_targets,
                  many_targets_sentence = f"Skipping {len(already_linked_targets)} already linked files or "
                                          f"directories.",
                  single_file_sentence = f"Skipping {len(already_linked_targets)} already linked file.",
                  single_directory_sentence = f"Skipping {len(already_linked_targets)} already linked directory.")

    skip_dotfiles(targets_list = corrupted_targets,
                  many_targets_sentence = f"Skipping {len(corrupted_targets)} files or directories because are links "
                                          f"to non-existing files or directories.",
                  single_file_sentence = f"Skipping {len(corrupted_targets)} file because is a link to a non-existing "
                                         f"file.",
                  single_directory_sentence = f"Skipping {len(corrupted_targets)} directory because is a link to a "
                                              f"non-existing directory.")

    skip_dotfiles(targets_list = wrong_existing_links,
                  many_targets_sentence = f"Skipping {len(wrong_existing_links)} files or directories because are "
                                          f"links to wrong existing files or directories.",
                  single_file_sentence = f"Skipping {len(wrong_existing_links)} file because is a link to a wrong "
                                         f"existing file.",
                  single_directory_sentence = f"Skipping {len(wrong_existing_links)} directory because is a link to a "
                                              f"wrong existing directory.")


def link_dotfiles(targets_list: dict[str, str],
                  settings_dir: str,
                  backup: bool,
                  many_targets_sentence: str,
                  single_file_sentence: str,
                  single_directory_sentence: str,
                  remove_settings: bool) -> None:
    if targets_list:
        n_targets = len(targets_list.keys())
        if n_targets > 1:
            for system_target_path, settings_target_path in targets_list.items():
                print_relationship(system_target_path = system_target_path,
                                   settings_target_path = settings_target_path,
                                   symbol = '-->')
            print_highlight(many_targets_sentence)
            proceed = ask_to_proceed()
        else:
            if os.path.isfile(list(targets_list.keys())[0]):
                print_relationship(system_target_path = list(targets_list.keys())[0],
                                   settings_target_path = list(targets_list.values())[0],
                                   symbol = '-->')
                print_highlight(single_file_sentence)
                proceed = ask_to_proceed()
            else:
                print_relationship(system_target_path = list(targets_list.keys())[0],
                                   settings_target_path = list(targets_list.values())[0],
                                   symbol = '-->')
                print_highlight(single_directory_sentence)
                proceed = ask_to_proceed()

        if proceed:
            for i, (system_target_path, settings_target_path) in enumerate(targets_list.items(), 1):
                try:
                    print_action(action_type = 'link',
                                 system_target_path = system_target_path,
                                 settings_target_path = settings_target_path)
                    print(f"Total ({i}/{n_targets})", end = '\r')

                    if remove_settings:
                        remove(path = settings_target_path)
                    copy(source = system_target_path, destination = settings_target_path)
                    change_parent_owner(source = system_target_path,
                                        destination = settings_target_path,
                                        settings_dir = settings_dir)
                    if os.path.isdir(system_target_path):
                        change_child_owner(source = system_target_path, destination = settings_target_path)
                    if backup:
                        backup_path = generate_backup_path(path = system_target_path)
                        os.rename(system_target_path, backup_path)
                    else:
                        remove(system_target_path)
                    os.symlink(settings_target_path, system_target_path)
                except Exception as e:
                    print_error(f"Error in linking {system_target_path}. Check {LOG_FILE_PATH} for more details.")
                    log_error(f'{e}')
        print('\n')



from argparse import Namespace
from configparser import ConfigParser
import os
from syndot.init_config import CONFIG_DIR_PATH
from syndot.utils.path import expand_home_path
from syndot.utils.map_file import read_map_file, write_map_file


DEFAULT_SETTINGS_DIR = '~/Settings'
MAP_TEMPLATE_PATH = os.path.join(CONFIG_DIR_PATH, 'templates', 'map.ini')
if not os.path.exists(CONFIG_DIR_PATH) or not os.path.exists(MAP_TEMPLATE_PATH):
    MAP_TEMPLATE_PATH = os.path.join(os.path.dirname(os.path.dirname(__file__)), '_templates', 'map.ini')


def init(args: Namespace) -> None:
    settings_dir = expand_home_path(args.path if args.path is not None else DEFAULT_SETTINGS_DIR)
    if os.path.exists(settings_dir):
        raise ValueError(f"Settings directory {settings_dir} already exists.")
    os.makedirs(settings_dir)

    config = read_map_file(MAP_TEMPLATE_PATH)
    config['Path']['settings_dir'] = settings_dir

    expand_default_target_path(config = config, target_category = 'files')
    expand_default_target_path(config = config, target_category = 'directories')

    write_map_file(map_file_path = os.path.join(settings_dir, 'map.ini'), config = config)


def expand_default_target_path(config: ConfigParser, target_category: str):
    targets = [expand_home_path(target) for target in config['Targets'][target_category].split()]
    targets = list(set(targets))
    targets.sort()
    config['Targets'][target_category] = '\n' + '\n'.join(targets)



from configparser import ConfigParser
import os
from syndot.init_config import CONFIG_DIR_PATH


COLORSCHEME_PATH = os.path.join(CONFIG_DIR_PATH, 'colorschemes', 'default.colorscheme')
if not os.path.exists(CONFIG_DIR_PATH) or not os.path.exists(COLORSCHEME_PATH):
    COLORSCHEME_PATH = os.path.join(os.path.dirname(os.path.dirname(__file__)), '_colorschemes', 'default.colorscheme')
color_config = ConfigParser()
color_config.read(COLORSCHEME_PATH)
COLOR_MAP = {'link': 'Color6',
             'symbol': 'Color0',
             'settings': 'Color1',
             'error': 'Color4'}


def get_ansi_color(color_type: str) -> str:
    color = color_config.get(section = COLOR_MAP[color_type], option = 'Color').replace(',', ';')
    return f"\x1b[38;2;{color}m"


class Color:
    LINK_COLOR = get_ansi_color('link')
    SYMBOL_COLOR = get_ansi_color('symbol')
    SETTINGS_COLOR = get_ansi_color('settings')
    ERROR_COLOR = get_ansi_color('error')
    BOLD_START_SEQUENCE = "\033[1m"
    BOLD_END_SEQUENCE = "\033[0m"
    COLOR_END_SEQUENCE = "\x1b[0m"

    @classmethod
    def link(cls, link: str) -> str:
        return cls.LINK_COLOR + link + cls.COLOR_END_SEQUENCE

    @classmethod
    def symbol(cls, symbol: str) -> str:
        return cls.SYMBOL_COLOR + symbol + cls.COLOR_END_SEQUENCE

    @classmethod
    def settings(cls, settings: str) -> str:
        return cls.SETTINGS_COLOR + settings + cls.COLOR_END_SEQUENCE

    @classmethod
    def error(cls, error: str) -> str:
        return cls.ERROR_COLOR + error + cls.COLOR_END_SEQUENCE

    @classmethod
    def highlight(cls, sentence: str) -> str:
        return cls.BOLD_START_SEQUENCE + sentence + cls.BOLD_END_SEQUENCE



from syndot.parser.add_parser import add_parser
from syndot.parser.diffuse_parser import diffuse_parser
from syndot.parser.init_parser import init_parser
from syndot.parser.link_parser import link_parser
from syndot.parser.list_parser import list_parser
from syndot.parser.remove_parser import remove_parser
from syndot.parser.unlink_parser import unlink_parser



from argparse import Namespace
import os
from syndot.init_config import LOG_FILE_PATH
from syndot.utils.commands import skip_dotfiles
from syndot.utils.file_actions import remove
from syndot.utils.logger import log_error
from syndot.utils.map_file import get_map_info, read_map_file
from syndot.utils.path import compose_target_paths
from syndot.utils.print_ import print_action, print_error, print_highlight, print_relationship
from syndot.utils.prompt import ask_to_proceed


def diffuse(args: Namespace) -> None:
    settings_dir, targets = get_map_info(config = read_map_file(map_file_path = args.mapfile), args = args)

    targets_to_be_diffused = {}
    already_existing_system = {}
    already_diffused_targets = []
    wrong_existing_links = {}
    missing_settings_targets = []
    settings_are_links = []

    print_highlight('Looking for files and directories to diffuse...')

    for target in targets:
        system_target_path, settings_target_path = compose_target_paths(settings_dir = settings_dir, target = target)
        if not os.path.islink(settings_target_path):
            if os.path.exists(settings_target_path):
                if not os.path.islink(system_target_path):
                    if not os.path.exists(system_target_path):
                        targets_to_be_diffused[system_target_path] = settings_target_path
                    else:
                        already_existing_system[system_target_path] = settings_target_path
                else:
                    if os.readlink(system_target_path) == settings_target_path:
                        already_diffused_targets.append(system_target_path)
                    else:
                        wrong_existing_links[system_target_path] = settings_target_path
            else:
                missing_settings_targets.append(settings_target_path)
        else:
            settings_are_links.append(settings_target_path)

    if not any([targets_to_be_diffused, already_existing_system, already_diffused_targets, wrong_existing_links,
                missing_settings_targets, settings_are_links]):
        print_highlight('No files or directories found to diffuse.')

    diffuse_dotfiles(targets_list = targets_to_be_diffused,
                     many_targets_sentence = f"Found {len(targets_to_be_diffused.keys())} files or directories to "
                                             f"diffuse.\nSymbolic links to these settings files and directories will "
                                             f"be created in the respective system directories.",
                     single_file_sentence = f"Found {len(targets_to_be_diffused.keys())} file to diffuse.\nA symbolic "
                                            f"link to this file will be created in the respective system directory.",
                     single_directory_sentence = f"Found {len(targets_to_be_diffused.keys())} directory to diffuse.\n"
                                                 f"A symbolic link to this directory will be created in the respective "
                                                 f"system directory.",
                     remove_system = False)

    diffuse_dotfiles(targets_list = already_existing_system,
                     many_targets_sentence = f"Found {len(already_existing_system.keys())} already existing system "
                                             f"files or directories.\nThey will be removed and replaced by symbolic "
                                             f"links to the respective files and directories in the settings "
                                             f"directory.",
                     single_file_sentence = f"Found {len(already_existing_system.keys())} already existing system "
                                            f"file.\nIt will be removed and replaced by a symbolic link to the "
                                            f"respective file in the settings directory.",
                     single_directory_sentence = f"Found {len(already_existing_system.keys())} already existing system "
                                                 f"directory.\nIt will be removed and replaced by a symbolic link to "
                                                 f"the respective directory in the settings directory.",
                     remove_system = True)

    diffuse_dotfiles(targets_list = wrong_existing_links,
                     many_targets_sentence = f"Found {len(wrong_existing_links.keys())} files or directories that are "
                                             f"links to wrong files or directories.\nThey will be removed and replaced "
                                             f"by symbolic links to the respective files and directories in the "
                                             f"settings directory.",
                     single_file_sentence = f"Found {len(wrong_existing_links.keys())} file that is a link to a wrong "
                                            f"file.\nIt will be removed and replaced by a symbolic link to the "
                                            f"respective file in the settings directory.",
                     single_directory_sentence = f"Found {len(wrong_existing_links.keys())} directory that is a link "
                                                 f"to a wrong directory.\nIt will be removed and replaced by a "
                                                 f"symbolic link to the respective directory in the settings "
                                                 f"directory.",
                     remove_system = True)

    skip_dotfiles(targets_list = already_diffused_targets,
                  many_targets_sentence = f"Skipping {len(already_diffused_targets)} already diffused files or "
                                          f"directories:",
                  single_file_sentence = f"Skipping {len(already_diffused_targets)} already diffused file:",
                  single_directory_sentence = f"Skipping {len(already_diffused_targets)} already diffused directory:")

    skip_dotfiles(targets_list = missing_settings_targets,
                  many_targets_sentence = f"Skipping {len(missing_settings_targets)} missing settings files or "
                                          f"directories:",
                  single_file_sentence = f"Skipping {len(missing_settings_targets)} missing settings file:",
                  single_directory_sentence = f"Skipping {len(missing_settings_targets)} missing settings directory:")

    skip_dotfiles(targets_list = settings_are_links,
                  many_targets_sentence = f"Skipping {len(settings_are_links)} settings files or directories because "
                                          f"are links:",
                  single_file_sentence = f"Skipping {len(settings_are_links)} settings file because is a link:",
                  single_directory_sentence = f"Skipping {len(settings_are_links)} settings directory because is a "
                                              f"link:")


def diffuse_dotfiles(targets_list: dict[str, str],
                     many_targets_sentence: str,
                     single_file_sentence: str,
                     single_directory_sentence: str,
                     remove_system: bool) -> None:
    if targets_list:
        n_targets = len(targets_list.keys())
        if n_targets > 1:
            for system_target_path, settings_target_path in targets_list.items():
                print_relationship(system_target_path = system_target_path,
                                   settings_target_path = settings_target_path,
                                   symbol = '<--')
            print_highlight(many_targets_sentence)
            proceed = ask_to_proceed()
        else:
            if os.path.isfile(list(targets_list.values())[0]):
                print_relationship(system_target_path = list(targets_list.keys())[0],
                                   settings_target_path = list(targets_list.values())[0],
                                   symbol = '<--')
                print_highlight(single_file_sentence)
                proceed = ask_to_proceed()
            else:
                print_relationship(system_target_path = list(targets_list.keys())[0],
                                   settings_target_path = list(targets_list.values())[0],
                                   symbol = '<--')
                print_highlight(single_directory_sentence)
                proceed = ask_to_proceed()

        if proceed:
            for i, (system_target_path, settings_target_path) in enumerate(targets_list.items(), 1):
                try:
                    print_action(action_type = 'diffuse',
                                 system_target_path = system_target_path,
                                 settings_target_path = settings_target_path)
                    print(f"Total ({i}/{n_targets})", end = '\r')

                    if remove_system:
                        remove(system_target_path)
                    if not os.path.exists(os.path.dirname(system_target_path)):
                        os.makedirs(os.path.dirname(system_target_path))
                    os.symlink(settings_target_path, system_target_path)
                except Exception as e:
                    print_error(f"Error in diffusing {system_target_path}. Check {LOG_FILE_PATH} for more details.")
                    log_error(f'{e}')
        print('\n')



from setuptools import setup
import subprocess


version = subprocess.run('git describe --tags'.split(), stdout = subprocess.PIPE).stdout.decode('utf-8').strip()


try:
    setup(version = version)
except Exception as e:
    if e.args[0].startswith("Invalid version: "):
        setup(version = 'v0.0.0')
    else:
        print(e)



from argparse import ArgumentParser, HelpFormatter, SUPPRESS, _SubParsersAction
from operator import attrgetter
from syndot.version import __version__


class GeneralFormatter(HelpFormatter):

    def __init__(self, prog):
        super().__init__(prog, max_help_position = 20, width = 80)

    def _iter_indented_subactions(self, action):
        if hasattr(action, '_get_subactions'):
            get_subactions = action._get_subactions
            if isinstance(action, _SubParsersAction):
                for subaction in sorted(get_subactions(), key = lambda x: x.dest):
                    yield subaction
            else:
                for subaction in get_subactions():
                    yield subaction

    def _format_action(self, action):
        result = super()._format_action(action)
        if isinstance(action, _SubParsersAction):
            return "%*s%s" % (self._current_indent, "", result.lstrip())
        return result

    def _format_action_invocation(self, action):
        if isinstance(action, _SubParsersAction):
            return ""
        return super()._format_action_invocation(action)


class CommandFormatter(HelpFormatter):

    def __init__(self, prog):
        super().__init__(prog, max_help_position = 40, width = 80)

    def add_arguments(self, actions):
        actions = sorted(actions, key = attrgetter('option_strings'))
        super(CommandFormatter, self).add_arguments(actions)

    def _format_action_invocation(self, action):
        if not action.option_strings or action.nargs == 0:
            return super()._format_action_invocation(action)
        default = self._get_default_metavar_for_optional(action)
        args_string = self._format_args(action, default)
        return ', '.join(action.option_strings) + ' ' + args_string


parser = ArgumentParser(prog = 'syndot',
                        usage = '%(prog)s {OPTION | COMMAND}',
                        description = 'Manage symlinks to dotfiles',
                        epilog = 'Config file path: ~/.config/syndot',
                        add_help = False,
                        formatter_class = GeneralFormatter)
parser._positionals.title = 'ARGUMENTS'
parser._optionals.title = 'OPTIONS'
parser.add_argument('-h', '--help',
                    action = 'help',
                    default = SUPPRESS,
                    help = 'Show this help message and exit')
parser.add_argument('-v', '--version',
                    action = 'version',
                    version = f'%(prog)s {__version__}',
                    help = 'Show program\'s version number and exit')
command_parser = parser.add_subparsers(dest = 'command',
                                       title = 'COMMANDS')



import logging
from syndot.init_config import LOG_FILE_PATH


def log_error(error_message: str):
    logging.basicConfig(filename = LOG_FILE_PATH,
                        level = logging.ERROR,
                        format = '%(asctime)s.%(msecs)03d %(levelname)s %(message)s',
                        datefmt = '%Y-%m-%d,%H:%M:%S')
    logger = logging.getLogger(__name__)
    logger.error(error_message)
